
from datetime import datetime, timedelta, timezone
from types import TracebackType
from typing import BinaryIO, Callable, List, Optional, Tuple, Type, TypeVar

from io_utils import Parser, read_fixed_point, read_u16, read_u64

def read_fixed(f: BinaryIO) -> float: # 16.16
	return read_fixed_point(f, 16, 16)

def read_fixed_version(f: BinaryIO) -> float: # Not the same as parse_fixed
	majorVersion = read_u16(f)
	minorVersion = read_u16(f)
	assert minorVersion in [0x0000, 0x1000, 0x5000], f"Invalid fixed minorVersion: {hex(minorVersion)}"

	return majorVersion + minorVersion/0xa000 # will need to change if there are ever any versions with 2 decimal digits

def read_F2DOT14(f: BinaryIO) -> float: # F2DOT14 (2.14)
	return read_fixed_point(f, 2, 14)

EPOCH = datetime(1904, 1, 1, tzinfo=timezone.utc)
def read_long_datetime(f: BinaryIO) -> datetime:
	return EPOCH+timedelta(seconds=read_u64(f))

T = TypeVar('T')

"""
The following `parse_at_...` functions all modify the tell of f, so it is recommended that you use `SaveTell` to wrap calling these.
"""

def parse_at_offset(f: BinaryIO, start_tell: int, offset: int, parser: Parser[T], *, zero_is_null:bool=True) -> T:
	"""
	Parses a `T` using `parser` at an offset of `offset` from `start_tell`. 
	
	If `zero_is_null` is True, then `offset` cannot be 0.
	Only set `zero_is_null` to False when you are sure that 0 is a valid offset
	"""
	if zero_is_null: assert offset != 0, f"Offset was NULL"
	f.seek(start_tell+offset)
	return parser(f)

def parse_at_optional_offset(f: BinaryIO, start_tell: int, offset: int, parser: Parser[T]) -> Optional[T]:
	"""
	Same as `parse_at_offset`, however if the offset is NULL (0), then None is returned.

	Should not be used when 0 is a valid offset to something.
	"""
	if offset == 0: return None
	return parse_at_offset(f, start_tell, offset, parser, zero_is_null=True)

def parse_at_offsets(f: BinaryIO, start_tell: int, offsets: List[int], parser: Parser[T], *, zero_is_null:bool=True) -> List[T]:
	return [parse_at_offset(f, start_tell, offset, parser, zero_is_null=zero_is_null) for offset in offsets]

def parse_at_optional_offsets(f: BinaryIO, start_tell: int, offsets: List[int], parser: Parser[T]) -> List[Optional[T]]:
	return [parse_at_optional_offset(f, start_tell, offset, parser) for offset in offsets]

def parse_at_offsets_using_length(f: BinaryIO, start_tell: int, offsets: List[int], parser: Callable[[BinaryIO, int, int], T], *, zero_is_null:bool=True) -> List[Optional[T]]:
	"""
	The length of the returned list will be one less than that of `offsets`, as the last offset is used to calculate the length of the final element.

	`parser` is of the form `(f: BinaryIO, index: int, length: int) -> T`
	"""

	elements: List[Optional[T]] = []
	for i, offset in enumerate(offsets[:-1]):
		length = offsets[i+1]-offset
		if length == 0:
			elements.append(None)
			continue

		elements.append(parse_at_offset(f, start_tell, offset, lambda f: parser(f, i, length), zero_is_null=zero_is_null))

	return elements

def parse_list_at_offset(f: BinaryIO, start_tell: int, offset: int, count: int, parser: Parser[T], *, zero_is_null:bool=True) -> List[T]:
	"""
	Parses a continuous list of `T`s with `count` elements.
	"""
	return parse_at_offset(f, start_tell, offset, lambda f: [parser(f) for _ in range(count)], zero_is_null=zero_is_null)

def parse_list_and_use_offsets_into(f: BinaryIO, start_tell: int, offsets: list[int], count: int, parser: Parser[T]) -> Tuple[List[T], List[T]]:
	elements = [(f.tell(), parser(f)) for _ in range(count)]
	elements_by_offset_into: List[T] = []
	for offset in offsets:
		for (offset_into, element) in elements:
			if offset + start_tell == offset_into:
				elements_by_offset_into.append(element)
				break
		else:
			assert False, (f"No element with offset {offset} into this list of elements", start_tell, offsets, elements)
	return elements_by_offset_into, [element for (_, element) in elements]

class SaveTell:
	"""
	A context manager that allows operations to be done on a BinaryIO without affecting the tell.
	"""
	def __init__(self, f: BinaryIO):
		self.f = f

	def __enter__(self) -> int:
		self.tell = self.f.tell()
		return self.tell

	def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], exc_tb: Optional[TracebackType]):
		self.f.seek(self.tell)


def null_if_zero(n: int) -> Optional[int]: return n if n else None
def nulls_if_zero(ns: List[int]) -> List[Optional[int]]: return list(map(null_if_zero, ns))