from io import SEEK_END
from typing import BinaryIO, Callable, Literal, TypeVar

def len_to_end(f: BinaryIO) -> int:
	curr_tell = f.tell()
	f.seek(0, SEEK_END)
	end_tell = f.tell()
	f.seek(curr_tell)
	return end_tell - curr_tell

def is_at_end(f: BinaryIO) -> bool:
	return len_to_end(f) == 0

ENDIANNESS: Literal['little', 'big'] = 'big'

def read_int_from_bytes(s: bytes, *, signed:bool=False) -> int: return int.from_bytes(s, ENDIANNESS, signed=signed)
def bytes_from_int(value: int, number: int, *, signed:bool=False) -> bytes: return value.to_bytes(number, ENDIANNESS, signed=signed)

def read_int(f: BinaryIO, number: int, *, signed:bool=False) -> int: return read_int_from_bytes(f.read(number), signed=signed)
def write_int(f: BinaryIO, value: int, number: int, signed:bool=False) -> int: return f.write(bytes_from_int(value, number, signed=signed))

def read_u64(f: BinaryIO) -> int: return read_int(f, 8)
def read_u32(f: BinaryIO) -> int: return read_int(f, 4)
def read_u24(f: BinaryIO) -> int: return read_int(f, 3)
def read_u16(f: BinaryIO) -> int: return read_int(f, 2)
def read_u8(f: BinaryIO) -> int: return read_int(f, 1)

def write_u16(f: BinaryIO, value: int) -> int: return write_int(f, value, 2)

def read_i32(f: BinaryIO) -> int: return read_int(f, 4, signed=True)
def read_i16(f: BinaryIO) -> int: return read_int(f, 2, signed=True)
def read_i8(f: BinaryIO) -> int: return read_int(f, 1, signed=True)

def read_utf8(f: BinaryIO, number: int) -> str: return f.read(number).decode(encoding='utf8')
def read_ascii(f: BinaryIO, number: int) -> str: return f.read(number).decode(encoding='ascii')

def read_fixed_point(f: BinaryIO, preradix_bits: int, postradix_bits:int, *, signed:bool=True) -> float:
	assert (preradix_bits+postradix_bits)%8 == 0
	raw = read_int(f, (preradix_bits+postradix_bits)//8, signed=signed)
	return raw/(1<<postradix_bits)

def read_pascal_string(f: BinaryIO) -> str:
	string_size = read_u8(f)
	pascal_string = read_ascii(f, string_size)
	return pascal_string

T = TypeVar('T')
Parser = Callable[[BinaryIO], T]